require 'test_helper'

describe "should character a CouchRest::Database" do

  it "create a couchrest doc" do
    
    doc = CouchRest::Document.new("_id" => 'test', "_rev" => 'hiho')
    assert !doc.new?
  end

  it "create an couchrest model" do
    
    doc = Testmodel.new({"_id" => 'test', "_rev" => 'hiho'}, :directly_set_attributes => true)
    assert !doc.new?
  end

  it "should character" do
    doc = create :a => 'A', :b => 'B'
    character = doc.to_character(Check)
    assert_equal "pong", character.ping
    assert_equal "A", character.a
    assert_equal "A", character["a"]
  end

  it "should update a character" do
    doc = create :a => 'A', :b => 'B', :characters => ['Check']
    character = doc.to_character(Check)
    assert_equal false, character.new?
    character.c = "C"
    character.save
    doc.reload
    assert_equal "ThingTank", doc["type"]
    assert_equal "B", doc["b"]
    assert_equal "C", doc["c"]
    assert_equal true, doc["check_saved"]
    assert_equal true, doc["check_updated"]
    assert doc["check_created"].nil?
  end

  it "should create a character, based on doc that did not have this character" do    
    doc = create :a => 'A', :b => 'B'
    character = doc.add_character(Check)
    assert character.new?, "character should be new when freshly added"
    character.c = "C"
    doc.save
    doc.reload
    assert_equal "ThingTank", doc["type"]
    assert_equal "B", doc["b"]
    assert_equal "C", doc["c"]
    assert_equal true, doc["check_saved"]
    assert_equal true, doc["check_created"]
    assert doc["check_updated"].nil?

    doc.reload
    character = doc.to_character(Check)
    assert_equal false, character.new?
  end

  it "should create a character, based on a new doc" do
    doc = new :a => 'A', :b => 'B', :characters => ['Check']
    character = doc.to_character(Check)
    assert character.new?
    character.c = "C"
    character.save
    assert_equal false, character.changed?
    doc.reload
    assert_equal "ThingTank", doc["type"]
    assert_equal "B", doc["b"]
    assert_equal "C", doc["c"]
    assert_equal true, doc["check_saved"]
    assert_equal true, doc["check_created"]
    assert doc["check_updated"].nil?

    doc.reload
    character = doc.to_character(Check)
    assert_equal false, character.new?
    assert_equal false, character.changed?
  end

  it "should not be valid if the requirements aren't met" do    
    doc = new 
    character = doc.to_character(Check)
    assert !character.valid?
  end

  it "should be able to access other characters and autosave the accessed characters if they are not saved by themselves" do    
    doc = new :a => 'A'
    character = doc.to_character(Check)
    character.access_other_character
    doc.save
    doc.reload
    assert_equal "I was here", doc["checker1"]
    assert_equal true, doc["characters"].include?(Checker.to_s)
    assert_equal true, doc["characters"].include?(Check.to_s)
  end

  it "should not be able to save properties we don't have" do
    doc = new :a => 'A'
    character = doc.to_character(Check)
    character.set_properties_we_don_t_have

    assert_raises RuntimeError do
      character.save
    end
  end

  it "should be able to save properties of the doc" do    
    doc = new :a => 'A'
    character = doc.to_character(Check)
    character.set_properties_of_doc
    character.save
    doc.reload
    assert_equal "got it", doc["ooops"]
  end

  it "should only destroy it's own properties" do
    doc = create :a => 'A', :c => 'C', :b => 'B', :shared => 'keep me too', :checker1 => 'keep me!', :characters => ['Check', 'Checker']
    character = doc.to_character(Check)
    doc.save
    doc.reload
    assert 'A', doc["a"]
    assert 'C', doc["c"]
    assert_equal "keep me!", doc["checker1"]
    assert_equal "keep me too", doc["shared"]
    assert doc["characters"].include?(Check.to_s)

    # now destroy Check
    character = doc.to_character(Check)
    character.destroy
    doc.save
    doc.reload
    assert_equal ['Checker'], doc["characters"]
    assert_equal "B", doc["b"]
    assert_equal "keep me!", doc["checker1"]
    assert_equal "keep me too", doc["shared"]
    assert doc["a"].nil?
  end

  it "should trying to make a character from a property" do
    doc = create :simple => { :a => 'A', :c => 'C', :b => 'B', :characters => ['Check']}, :checker1 => 'keep me!', :characters => ['Checker']
    character = doc.to_character(Check,'simple')
    character.c = "BC"
    doc.save
    assert_equal "BC", doc["simple"]["c"]
    assert doc["simple"]["check_saved"]
  end

  it "should trying to make a array of characters from a property" do
    doc = create :simple => [{ :a => 'A1', :c => 'C1', :b => 'B1', :characters => ['Check']}, { :a => 'A2', :c => 'C2', :b => 'B2', :characters => ['Check']}], :checker1 => 'keep me!', :characters => ['Checker']
    characters = doc.to_character(Check, 'simple')
    characters.each do |fk|
      fk.c = fk.c.downcase
    end
       
    doc.save
    assert_equal "c1", doc["simple"].first["c"]
    assert doc["simple"].first["check_saved"]
    assert_equal "c2", doc["simple"].last["c"]
    assert doc["simple"].last["check_saved"]
  end


  it "should also possible to use a character as property of a character" do
    checks = [{ :a => 'A1', :c => 'C1', :b => 'B1', :characters => ['Check']}, { :a => 'A2', :c => 'C2', :b => 'B2', :characters => ['Check']}]
    doc = create :checker1 => checks, :characters => ['Checker']
    character = doc.to_character(Checker)

    character.to_character(Check, 'checker1').each do |fk|
      fk.c = fk.c.downcase
    end

    doc.save
    assert_equal "c1", doc["checker1"].first["c"]
    assert_equal "c2", doc["checker1"].last["c"]
  end

  it "should also possible to use a character as property of anotherone" do
    doc = create :simple => { :a => 'AAAA', :c => {:checker1 => 'modify me!!!!!', :characters => ['Checker']}, :b => 'B', :characters => ['Check']}
    character1 = doc.to_character(Check, 'simple')
    character2 = character1.to_character(Checker, 'c')
    character2.checker1 = "modified!!!!"
    doc.save
    assert_equal "modified!!!!", doc["simple"]["c"]["checker1"]
  end

  it "should also possible to use a character as property of anotherone (array)" do
    doc = create :simple => { :a => 'AAAA', :c => [{:checker1 => 'mod1', :characters => ['Checker']},{:checker1 => 'mod2', :characters => ['Checker']}], :b => 'B', :characters => ['Check']}
    character1 = doc.to_character(Check, 'simple')
    characters = character1.to_character(Checker, 'c')
    characters.each do |fk|
      fk.checker1 = fk.checker1.upcase
    end

    doc.save
    assert_equal "MOD1", doc["simple"]["c"].first["checker1"]
    assert_equal "MOD2", doc["simple"]["c"].last["checker1"]
  end

  it "should also possible to use a character as property of anotherone (array.save)" do
    doc = create :simple => { :a => 'AAAA', :c => [{:checker1 => 'mod1', :characters => ['Checker']},{:checker1 => 'mod2', :characters => ['Checker']}], :b => 'B', :characters => ['Check']}
    character1 = doc.to_character(Check, 'simple')
    characters = character1.to_character(Checker, 'c')
    characters.each do |fk|
      fk.checker1 = fk.checker1.upcase
    end

    # saves only the first
    characters.first.save
    assert_equal "MOD1", doc["simple"]["c"].first["checker1"]
    assert_equal "mod2", doc["simple"]["c"].last["checker1"]
    
    # saves all
    doc.save
    assert_equal "MOD1", doc["simple"]["c"].first["checker1"]
    assert_equal "MOD2", doc["simple"]["c"].last["checker1"]
  end

  it "should also possible to use a character as property of anotherone (destroy hook)" do
    doc = create :simple => { :a => 'AAAA', :c => {:checker1 => 'modify me!!!!!', :characters => ['Checker'], :a => 'A'}, :b => 'B', :characters => ['Check', 'Checker']}
    character1 = doc.to_character(Check, 'simple')
    character2 = character1.to_character(Checker, 'c')
    character2.destroy
    doc.save
    assert_equal 'A', doc["simple"]["c"]["a"]
    assert doc["simple"]["c"]["checker1"].nil?
  end


  it "should also possible to use a character as property of anotherone (array. trigger destroy hook)" do
    doc = create :simple => { :a => 'AAAA', :c => [{:checker1 => 'mod1', :a => 'A', :characters => ['Checker']},{:checker1 => 'mod2', :characters => ['Checker']}], :b => 'B', :characters => ['Check']}
    character1 = doc.to_character(Check, 'simple')
    f = character1.first_character(Checker, 'c')

    assert_equal 'mod1', doc["simple"]["c"].first["checker1"]
    assert_equal 'A', doc["simple"]["c"].first["a"]


    assert_equal 'mod1', f.checker1
    f.destroy
    assert_equal nil, f.checker1
    assert_equal 'A', f._character_doc["a"]
    doc.save
    doc.reload

    assert_equal nil, doc["simple"]["c"].first["checker1"]
    assert_equal 'A', doc["simple"]["c"].first["a"]
  end


  it "should allow adding of a character as property with properties directly set" do
    doc = create 
    doc.add_character(Check, 'check') do |c|
      c.a = 'A'
    end
  
    assert_equal 'A', doc["check"]["a"]
  end

  it "should allow adding of a character as property without properties directly set" do
    doc = create 
    doc.add_character(Check, 'check')
    
    doc.last_character(Check, 'check') do |c|
      c.a = 'A'
    end

    doc.last_character(Check, 'check') do |c|
      c.c = 'C'
    end

  
    assert_equal 'A', doc["check"]["a"]
    assert_equal 'C', doc["check"]["c"]
  end

  it "should allow adding of a character as property multiple times transforming into an array by the way" do
    doc = create 
    doc.add_character(Check, 'check') do |c|
      c.a = 'A1'
    end

    assert_equal 'A1', doc["check"]["a"]
    
    doc.add_character(Check, 'check') do |c|
      c.a = 'A2'
    end

    doc.add_character(Check, 'check') do |c|
      c.a = 'A3'
    end

    assert_equal 'A1', doc["check"].first["a"]
    assert_equal 'A2', doc["check"][1]["a"]
    assert_equal 'A3', doc["check"].last["a"]
  end

  it "should allow adding of a character of a character as property with properties directly set 1" do
    doc = create 
    doc.add_character(Check, 'check') do |c|
      c.a = 'A'
    end

    character = doc.to_character(Check, 'check')

    character.add_character(Checker, 'sub') do |ck|
      ck.checker1 = 'hiho'
    end

    assert_equal 'hiho', doc["check"]["sub"]["checker1"]

    character.add_character(Checker, 'sub') do |ck|
      ck.checker1 = 'hoho'
    end

  
    assert_equal 'A', doc["check"]["a"]
    assert_equal 'hiho', doc["check"]["sub"].first["checker1"]
    assert_equal 'hoho', doc["check"]["sub"].last["checker1"]
  end


  it "should allow adding of a character of a character as property with properties directly set 2" do
    doc = create 
    doc.add_character(Check, 'check') do |c|
      c.a = 'A1'
    end

    doc.add_character(Check, 'check') do |c|
      c.a = 'A2'
    end


    character1 = doc.first_character(Check, 'check')

    character1.add_character(Checker, 'sub') do |ck|
      ck.checker1 = 'A1sub1'
    end

    assert_equal 'A1sub1', doc["check"].first["sub"]["checker1"]

    character1.add_character(Checker, 'sub') do |ck|
      ck.checker1 = 'A1sub2'
    end

    character2 = doc.last_character(Check, 'check')

    character2.add_character(Checker, 'sub') do |ck|
      ck.checker1 = 'A2sub1'
    end

    assert_equal 'A2sub1', doc["check"].last["sub"]["checker1"]

    character2.add_character(Checker, 'sub') do |ck|
      ck.checker1 = 'A2sub2'
    end


    assert_equal 'A1sub1', doc["check"].first["sub"].first["checker1"]
    assert_equal 'A1sub2', doc["check"].first["sub"].last["checker1"]
    assert_equal 'A2sub1', doc["check"].last["sub"].first["checker1"]
    assert_equal 'A2sub2', doc["check"].last["sub"].last["checker1"]

  
    assert_equal 'A1', doc["check"].first["a"]
    assert_equal 'A2', doc["check"].last["a"]
  end


  it "should allow adding of a character as property without properties directly set" do
    doc = create 
    doc.add_character(Check, 'check') do |c|
      c.a = 'A'
    end
    
    character1 = doc.last_character(Check, 'check')

    character1.add_character(Checker, 'sub')

    character1.last_character(Checker, 'sub') do |ck|
      ck.checker1 = 'A1sub1'
    end

    assert_equal 'A', doc["check"]["a"]
    assert_equal 'A1sub1', doc["check"]["sub"]["checker1"]
  end

  it "should not save the doc when saving a character, but flush its changes to the doc" do
    doc = create :test => 'hiho'
    doc['test'] = 'hoho'

    character = doc.add_character(Check)

    character.a = 'A'

    assert_equal nil, doc['a']
    assert_equal doc, character._doc
    character.to_doc
    assert_equal 'A', doc['a']
    assert_equal 'hoho', doc['test']
    assert_equal 'hiho', db_load(doc.id)['test']

  end

end
