require 'test_helper'

class TestC < ThingTank::Character
  property :propC
  validates_presence_of :propC
end

class TestB < ThingTank::Character
  property :propB
  validates_presence_of :propB
end

class TestA < ThingTank::Character
  property :prop1
  property :prop2
  property :foo1
  property :foo2
  property :doc
  property :inline_doc

  validates :prop1, :class => Array
  validates :prop2, :character => TestB
  validates :foo1, :class => [Integer, String]
  validates :foo2, :character => [TestB, TestC]

  validates :doc, :document => true
  validates :inline_doc, :document => {:inline => true}
end

describe "a classvalidator" do
  it "should validate correctly single class" do
    assert_equal true, TestA.new(:prop1 => ["Something"]).valid?
    assert_equal false, TestA.new(:prop1 => "Something").valid?
  end

  it "should validate correctly multiple classes" do
    assert_equal true, TestA.new(:foo1 => "Something").valid?
    assert_equal true, TestA.new(:foo1 => 5).valid?
    assert_equal false, TestA.new(:foo1 => {:hi => "ho"}).valid?
  end
end

describe "a charactervalidator" do
  it "should validate correctly single valid character" do
    doc = ThingTank.create
    doc.add_character TestA do |c|
      c.prop2 = {:propB => "bla"}
    end
    assert_equal true, doc.valid?
  end

  it "should validate correctly single invalid character" do
    doc = ThingTank.create
    doc.add_character TestA do |c|
      c.prop2 = {:prop5 => "bla"}
    end
    assert_equal false, doc.as(TestA).valid?
  end

  it "should validate correctly multiple valid characters" do
    doc = ThingTank.create
    doc.add_character TestA do |c|
      c.foo2 = {:propB => "bla"}
    end
    assert_equal true, doc.as(TestA).valid?

    doc2 = ThingTank.create
    doc2.add_character TestA do |c|
      c.foo2 = {:propC => "bla"}
    end
    assert_equal true, doc2.as(TestA).valid?
  end

  it "should validate correctly multiple invalid characters" do
    doc = ThingTank.create
    doc.add_character TestA do |c|
      c.foo2 = {:prop5 => "bla"}
    end
    assert_equal false, doc.as(TestA).valid?
  end
end


describe "a charactervalidator and an array of values" do
  it "should validate correctly single valid character" do
    doc = ThingTank.create
    doc.add_character TestA do |c|
      c.prop2 = [{:propB => "bla"}, {:propB => "blub"}]
    end
    assert_equal true, doc.valid?
  end

  it "should validate correctly single invalid character" do
    doc = ThingTank.create
    doc.add_character TestA do |c|
      c.prop2 = [{:prop5 => "bla"}, {:propB => "bla"}]
    end
    assert_equal false, doc.as(TestA).valid?
  end

  it "should validate correctly multiple valid characters" do
    doc = ThingTank.create
    doc.add_character TestA do |c|
      c.foo2 = [{:propB => "bla"}, {:propB => "blubb"}]
    end
    assert_equal true, doc.as(TestA).valid?

    doc2 = ThingTank.create
    doc2.add_character TestA do |c|
      c.foo2 = [{:propC => "bla"}, {:propC => "bla"}]
    end
    assert_equal true, doc2.as(TestA).valid?
  end

  it "should validate correctly multiple invalid characters" do
    doc = ThingTank.create
    doc.add_character TestA do |c|
      c.foo2 = [{:prop5 => "bla"}, {:propB => "bla"}]
    end
    assert_equal false, doc.as(TestA).valid?
  end
end

describe "a doc validator" do
  it "should validate a doc id" do
    doc1 = ThingTank.create :a => 'b'
    doc1.save
    doc2 = ThingTank.create
    doc2.add_character TestA do |c|
      c.doc = doc1.to_sym
    end
    assert_equal true, doc2.as(TestA).valid?

    doc3 = ThingTank.create
    doc3.add_character TestA do |c|
      c.doc = doc1.id
    end
    assert_equal false, doc3.as(TestA).valid?

    doc4 = ThingTank.create
    doc4.add_character TestA do |c|
      c.doc = "!#3" + doc1.id
    end
    assert_equal false, doc4.as(TestA).valid?

    doc5 = ThingTank.create
    doc5.add_character TestA do |c|
      c.doc = doc1
    end
    assert_equal true, doc5.as(TestA).valid?
  end

  it "should validate an inline doc" do
    doc1 = ThingTank.new :a => 'b'

    doc2 = ThingTank.create
    doc2.add_character TestA do |c|
      c.inline_doc = doc1.to_character_hash(true)
    end
    assert_equal true, doc2.as(TestA).valid?

    doc3 = ThingTank.create
    doc3.add_character TestA do |c|
      c.inline_doc = doc1
    end

    character = doc3.as(TestA)
    assert_equal true, doc3.as(TestA).valid?

    doc4 = ThingTank.create
    doc4.add_character TestA do |c|
      c.inline_doc = "test"
    end

    character = doc4.as(TestA)
    assert_equal false, doc4.as(TestA).valid?
  end


  it "should validate multiple doc ids" do
    doc0 = ThingTank.create :a => 'b'
    doc0.save
    doc1 = ThingTank.create :c => 'd'
    doc1.save
    doc2 = ThingTank.create
    doc2.add_character TestA do |c|
      c.doc = [doc1.to_sym, doc0.to_sym]
    end
    assert_equal true, doc2.as(TestA).valid?

    doc3 = ThingTank.create
    doc3.add_character TestA do |c|
      c.doc = [doc1.id, doc0.to_sym]
    end
    assert_equal false, doc3.as(TestA).valid?

    doc4 = ThingTank.create
    doc4.add_character TestA do |c|
      c.doc = ["!#3" + doc1.id, doc0.to_sym]
    end
    assert_equal false, doc4.as(TestA).valid?

    doc5 = ThingTank.create
    doc5.add_character TestA do |c|
      c.doc = [doc1, doc0]
    end
    assert_equal true, doc5.as(TestA).valid?
  end

  it "should validate an multiple inline doc" do
    doc0 = ThingTank.new :a => 'b'
    doc1 = ThingTank.new :c => 'd'

    doc2 = ThingTank.create
    doc2.add_character TestA do |c|
      c.inline_doc = [doc1.to_character_hash(true), doc0.to_character_hash(true)]
    end
    assert_equal true, doc2.as(TestA).valid?

    doc3 = ThingTank.create
    doc3.add_character TestA do |c|
      c.inline_doc = [doc1, doc0]
    end

    character = doc3.as(TestA)
    assert_equal true, doc3.as(TestA).valid?

    doc4 = ThingTank.create
    doc4.add_character TestA do |c|
      c.inline_doc = ["test", doc1]
    end

    character = doc4.as(TestA)
    assert_equal false, doc4.as(TestA).valid?
  end


end